package org.processmining.processtree;

import java.util.Collection;
import java.util.Iterator;
import java.util.UUID;

public interface ProcessTree extends ProcessTreeElement {

	/**
	 * Possible types of node. For each final subclass of node there is a
	 * corresponding type defined here.
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static enum Type {
		XOR("Xor", Block.Xor.class), //
		OR("Or", Block.Or.class), //
		AND("And", Block.And.class), //
		SEQ("Sequence", Block.Seq.class), //
		DEF("Deferred choice", Block.Def.class), //
		LOOPXOR("Xor loop", Block.XorLoop.class), //
		LOOPDEF("Deferred loop", Block.DefLoop.class), //
		PLACEHOLDER("Placeholder", Block.PlaceHolder.class), //
		AUTOTASK("Automatic task", Task.Automatic.class), //
		MANTASK("Manual task", Task.Manual.class), //
		TIMEOUT("Timeout event", Event.TimeOut.class), //
		MESSAGE("Message event", Event.Message.class); //

		private final String name;
		private final Class<? extends Node> type;

		Type(String name, Class<? extends Node> type) {
			this.name = name;
			this.type = type;
		}

		public String toString() {
			return name;
		}

		public Class<? extends Node> getType() {
			return type;
		}

		public static Type getTypeOf(Node node) {
			for (Type t : Type.values()) {
				if (t.getType().isAssignableFrom(node.getClass())) {
					return t;
				}
			}
			throw new RuntimeException("Unknown node type");
		}
	}

	/**
	 * Returns the root of the process tree
	 * 
	 * @return
	 */
	public Node getRoot();

	/**
	 * Sets the root of this process tree.
	 * 
	 * @param root
	 */
	public void setRoot(Node root);

	/**
	 * Returns an immutable collection of nodes in this tree. No guarantees are
	 * given about the ordering of nodes, not even about consistency in ordering
	 * between calls.
	 * 
	 * @return
	 */
	public Collection<Node> getNodes();

	/**
	 * Returns an immutable collection of edges in this tree. No guarantees are
	 * given about the ordering of edges, not even about consistency in ordering
	 * between calls.
	 * 
	 * @return
	 */
	public Collection<Edge> getEdges();

	/**
	 * Returns an immutable collection of variables used in this tree. No
	 * guarantees are given about the ordering of variables, not even about
	 * consistency in ordering between calls.
	 * 
	 * @return
	 */
	public Collection<Variable> getVariables();
	
	/**
	 * Returns an immutable collection of expressions used in this tree. No
	 * guarantees are given about the ordering of expressions, not even about
	 * consistency in ordering between calls.
	 * 
	 * @return
	 */
	public Collection<Expression> getExpressions();

	/**
	 * Returns an immutable collection of originators used in this tree. No
	 * guarantees are given about the ordering of originators, not even about
	 * consistency in ordering between calls.
	 * 
	 * @return
	 */
	public Collection<Originator> getOriginators();

	/**
	 * adds a node to the tree. Returns true if the tree changed due to the
	 * addition of this node. Returns false if the node was already in the tree.
	 * 
	 * 
	 * @return
	 */
	public boolean addNode(Node node);
	
	/**
	 * removes a node from the tree. Returns true if the tree changed due to the
	 * removal of this node. Returns false if the node was not in the tree.
	 * 
	 * 
	 * @return
	 */
	public boolean removeNode(Node node);

	/**
	 * adds a variable to the tree. Returns true if the tree changed due to the
	 * addition of this variable. Returns false if the variable was already in
	 * the tree.
	 * 
	 * @return
	 */
	public boolean addVariable(Variable variable);
	
	/**
	 * removes a variable from the tree. Returns true if the tree changed due to the
	 * removal of this variable. Returns false if the variable was not in the tree
	 * 
	 * @param variable
	 * @return
	 */
	public boolean removeVariable(Variable variable);

	/**
	 * adds an originator to the tree. Returns true if the tree changed due to
	 * the addition of this originator. Returns false if the originator was
	 * already in the tree.
	 * 
	 * @return
	 */
	public boolean addOriginator(Originator originator);
	
	/**
	 * removes an originator from the tree. Returns true if the tree changed due to the
	 * removal of this originator. Returns false if the originator was not in the tree
	 * 
	 * @param originator
	 * @return
	 */
	public boolean removeOriginator(Originator originator);

	/**
	 * adds an edge to the tree. Returns true if the tree changed due to the
	 * addition of this edge. Returns false if the edge was already in the tree.
	 * 
	 * @return
	 */
	public boolean addEdge(Edge edge);
	
	/**
	 * removes an edge from the tree. Returns true if the tree changed due to the
	 * removal of this edge. Returns false if the edge was not in the tree.
	 * 
	 * @return
	 */
	public boolean removeEdge(Edge edge);
	
	
	/**
	 * adds an expression to the tree. Returns true if the tree changed due to the 
	 * addition of this expression. Returns false if the expression was already in the tree.
	 * 
	 * @param expression
	 * @return
	 */
	public boolean addExpression(Expression expression);
	
	/**
	 * removes an expression from the tree. Returns true if the tree changed due to the
	 * removal of this expression. Returns false if the expression was not in the tree.
	 * 
	 * @param expression
	 * @return
	 */
	public boolean removeExpression(Expression expression);
	

	/**
	 * returns an iterator over the nodes implementing the given type.
	 * 
	 * @param <N>
	 *            the type returned by type.getType()
	 * @param n
	 * @return
	 */
	public <N extends Node> Iterator<N> iterator(Type type);

	/**
	 * Returns the type of a node expressed in the Type enumeration
	 * 
	 * @param node
	 * @return
	 */
	public Type getType(Node node);

	/**
	 * Returns the number of nodes in the tree
	 * 
	 * @return
	 */
	public int size();

	/**
	 * Returns true if and only if this processtree is a proper tree, i.e. when
	 * visiting all nodes depth-first there are no two visits to the same node.
	 * 
	 * @return
	 */
	public boolean isTree();

	/**
	 * Returns true if and only if this processtree is a directed acyclic graph.
	 * This should always be the case but is not necessarily enforces by
	 * implementing classes
	 * 
	 * @return
	 */
	public boolean isDag();	
	
	/**
	 * Return a tree representation of this processtree (i.e., the DAG is unfolded)
	 * 
	 * @return
	 */
	public ProcessTree toTree();
	
	/**
	 * Returns true if and only if this processtree has configuration options
	 * @return
	 */
	public boolean isConfigurable();
	
	/**
	 * Return the node with the given ID
	 * @param id
	 * @return
	 */
	public Node getNode(UUID id);
	
	/**
	 * Return the originator with the given ID
	 * @param id
	 * @return
	 */
	public Originator getOriginator(UUID id);
	
	/**
	 * Return the edge with the given ID
	 * @param id
	 * @return
	 */
	public Edge getEdge(UUID id);
	
	/**
	 * Return the variable with the given ID
	 * @param id
	 * @return
	 */
	public Variable getVariable(UUID id);
	
	/**
	 * Return the expression with the given ID
	 * @param id
	 * @return
	 */
	public Expression getExpression(UUID id);
	
	/**
	 * Return all the process tree elements
	 * @return
	 */
	public Collection<ProcessTreeElement> getElements();
	
	/**
	 * Return the process tree element with the given ID
	 * @param id
	 * @return
	 */
	public ProcessTreeElement getElement(UUID id);
	
	
	/**
	 * Return a string representing a node
	 * @param node
	 * @return
	 */
	public String toString(Node node);
}
