package org.processmining.processtree.configuration.controlflow;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.processmining.processtree.Block.PlaceHolder;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;

public interface ControlFlowConfiguration {
	/**
	 * adds a blocked edge
	 * @param e
	 */
	void addBlockedEdge(Edge e);

	/**
	 * adds a blocked edge
	 * @param e
	 */
	void addBlockedEdge(UUID e);

	/**
	 * adds a number of blocked edges
	 * @param ids
	 */
	void addBlockedEdges(Set<UUID> ids);
	
	/** 
	 * adds a non-blocked edge
	 * @param e
	 */
	void addNonBlockedEdge(Edge e);

	/**
	 * adds a non-blocked edge
	 * @param e
	 */
	void addNonBlockedEdge(UUID e);
	
	/**
	 * adds a number of non blocked edges
	 * @param ids
	 */
	void addNonBlockedEdges(Set<UUID> ids);

	/**
	 * adds a hidden edge
	 * @param e
	 */
	void addHiddenEdge(Edge e);

	/**
	 * adds a hidden edge
	 * @param e
	 */
	void addHiddenEdge(UUID e);
	
	/**
	 * adds a number of hidden edges
	 * @param ids
	 */
	void addHiddenEdges(Set<UUID> ids);

	/**
	 * adds a non-hidden edge
	 * @param e
	 */
	void addNonHiddenEdge(Edge e);

	/**
	 * adds a non-hidden edge
	 * @param e
	 */
	void addNonHiddenEdge(UUID e);
	
	/**
	 * adds a number of non hidden edges
	 * @param ids
	 */
	void addNonHiddenEdges(Set<UUID> ids);

	/**
	 * sets a replacement option for a placeholder node
	 * @param placeholder
	 * @param selectedChild
	 */
	void setSelectedReplacement(PlaceHolder placeholder, Node selectedChild);

	/**
	 * sets a replacement option for a placeholder node
	 * @param placeholder
	 * @param selectedChild
	 */
	void setSelectedReplacement(UUID placeholder, UUID selectedChild);

	/**
	 * sets a replacement option as not selected for a placeholder node
	 * @param placeholder
	 * @param nonSelectedChild
	 */
	void addNonSelectedReplacement(PlaceHolder placeholder, Node nonSelectedChild);

	/**
	 * sets a replacement option as not selected for a placeholder node
	 * @param placeholder
	 * @param nonSelectedChild
	 */
	void addNonSelectedReplacement(UUID placeholder, UUID nonSelectedChild);

	/**
	 * sets a changeable node to a specific type
	 * @param node
	 * @param type
	 */
	void setChangeableNode(Node node, Type type);

	/**
	 * sets a changeable node to a specific type
	 * @param node
	 * @param type
	 */
	void setChangeableNode(UUID node, Type type);

	/**
	 * removes a changeable node
	 * @param node
	 */
	void addNonChangeableNode(Node node);

	/**
	 * removes a changeable node
	 * @param node
	 */
	void addNonChangeableNode(UUID node);

	/**
	 * returns if an edge is blocked
	 * @param e
	 * @return
	 */
	boolean isBlocked(Edge e);
	
	/**
	 * returns if an edge is not blocked
	 * @param e
	 * @return
	 */
	boolean isNonBlocked(Edge e);

	/**
	 * returns if an edge is hidden
	 * @param e
	 * @return
	 */
	boolean isHidden(Edge e);

	/**
	 * returns if an edge is not hidden
	 * @param e
	 * @return
	 */
	boolean isNonHidden(Edge e);

	/**
	 * returns the set of blocked edges
	 * @return
	 */
	Set<UUID> getBlockedEdges();

	/**
	 * returns the set of non-blocked edges
	 * @return
	 */
	Set<UUID> getNonBlockedEdges();

	/**
	 * returns the set of hidden edges
	 * @return
	 */
	Set<UUID> getHiddenEdges();

	/**
	 * returns the set of non-hidden edges
	 * @return
	 */
	Set<UUID> getNonHiddenEdges();

	/**
	 * returns the selected replacement options
	 * @return
	 */
	Map<UUID, UUID> getSelectedReplacements();

	/**
	 * returns the non-selected replacement options
	 * @return
	 */
	Map<UUID, List<UUID>> getNonSelectedReplacements();

	/**
	 * returns the replaced nodes
	 * @return
	 */
	Map<UUID, Type> getReplacedNodes();

	/**
	 * returns the non-replaced nodes
	 * @return
	 */
	Set<UUID> getNonReplacedNodes();
	
	/**
	 * Special version of the toString method where the UUIDs are searched for
	 * in the provided process tree (NULL allowed) and the appropriate toString
	 * method is called for a more intuitive string result
	 * 
	 * @param pt
	 * @return
	 */
	String toString(ProcessTree pt);
}
