package org.processmining.processtree.configuration.originator.impl;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.processmining.framework.util.Pair;
import org.processmining.processtree.Node;
import org.processmining.processtree.Originator;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task.Manual;
import org.processmining.processtree.configuration.ConfigurationAuxiliary;
import org.processmining.processtree.configuration.originator.OriginatorConfiguration;
import org.processmining.processtree.configuration.originator.OriginatorConfigurationIterator;
import org.processmining.processtree.impl.ProcessTreeImpl;

public class OriginatorConfigurationIteratorImpl implements OriginatorConfigurationIterator{

	private ProcessTree tree;
	public BitSet currentConfiguration;
	private List<UUID> manualTasks;
	public Map<UUID, List<UUID>> remOrgs;
	public BitSet lastReturnedConfiguration = null;
	private int bitLength;
	private BitSet minConfiguration = null;
	private BitSet maxConfiguration = null;

	public OriginatorConfigurationIteratorImpl(ProcessTree tree) throws Exception{
		this(tree, null, null);
	}
	
	public OriginatorConfigurationIteratorImpl(ProcessTree tree, Object minConfiguration, Object maxConfiguration){
		this.tree = new ProcessTreeImpl(tree);
		this.minConfiguration = (BitSet) minConfiguration;
		this.maxConfiguration = (BitSet) maxConfiguration;
		fillRemoved();		
		this.bitLength = getBitLength();
		sort();
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
	}
	
	private void fillRemoved(){
		remOrgs = new HashMap<UUID, List<UUID>>();
		manualTasks = new ArrayList<UUID>();
		for(Node n: tree.getNodes()){
			if(n instanceof Manual){
				if(!(((Manual)n).getRemovableOriginators().isEmpty())){
					// lets make the ArrayList
					List<UUID> orgs = new ArrayList<UUID>();
					for(Originator org: ((Manual)n).getRemovableOriginators()){
						orgs.add(org.getID());
					}
					remOrgs.put(n.getID(), orgs);
					manualTasks.add(n.getID());
				}
			}
		}
	}
	
	private int getBitLength(){
		int bitLength = 0;
		for(UUID id: remOrgs.keySet()){
			bitLength += remOrgs.get(id).size();
		}
		return bitLength;
	}
	
	private void sort(){
		Collections.sort(manualTasks);
		for(UUID id: manualTasks){
			Collections.sort(remOrgs.get(id));
		}
	}
	
	public OriginatorConfiguration nextConfiguration() throws Exception{
		OriginatorConfiguration conf = new OriginatorConfigurationImpl();
		if(lastReturnedConfiguration == null){
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(currentConfiguration.equals(lastReturnedConfiguration) && !currentConfiguration.get(currentConfiguration.length() - 2)){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		while((!isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration))){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			conf = new OriginatorConfigurationImpl();
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && !(currentConfiguration.equals(lastReturnedConfiguration)) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration)){
			lastReturnedConfiguration = (BitSet) currentConfiguration.clone();
			return conf;
		}
		else{
			return null;
		}
	}
	
	public boolean hasNext() throws Exception{
		BitSet currentC = (BitSet) currentConfiguration.clone();
		BitSet lastRetConf = null;
		if(lastReturnedConfiguration != null){
			lastRetConf = (BitSet) lastReturnedConfiguration.clone();
		}
		if(nextConfiguration() != null){
			currentConfiguration = currentC;
			lastReturnedConfiguration = lastRetConf;
			return true;
		}
		else{
			return false;
		}
	}
	
	public OriginatorConfiguration getNext() throws Exception {
		if(hasNext()){
			return nextConfiguration();
		}
		throw new NullPointerException();
	}
	
	public void reset(){
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
		lastReturnedConfiguration = null;
	}

	private void fillConfiguration(OriginatorConfiguration conf){
		// TODO: this can be done more efficient
		// note the -2 due to the control bits
		for(int i = 0; i < currentConfiguration.length() - 2; i++){
			// we need to have dynamic reduction on the length of the lists, i.e., we need to lookup to which manual task and originator this belongs
			int lowerboundOfInterval = 0;
			for(UUID id: manualTasks){
				if(lowerboundOfInterval <= i && i < lowerboundOfInterval + remOrgs.get(id).size()){
					if(currentConfiguration.get(i)){
						conf.addRemovedOriginator(id, remOrgs.get(id).get(i - lowerboundOfInterval));
					}
				}
				else{
					lowerboundOfInterval += remOrgs.get(id).size();
				}
			}
		}
	}
		
	protected static void makeComplete(OriginatorConfiguration conf, ProcessTree tree){
		for(Node n: tree.getNodes()){
			if(n instanceof Manual){
				for(Originator org: ((Manual)n).getRemovableOriginators()){
					if(!conf.getRemovedOriginator(n.getID()).contains(org.getID())){
						conf.addNonRemovedOriginator(n.getID(), org.getID());
					}
				}
			}
		}
	}
	
	private boolean isConsistentComplete(OriginatorConfiguration conf, ProcessTree tree){
		for(UUID id: manualTasks){
			if(ConfigurationAuxiliary.overlap(conf.getRemovedOriginator(id), conf.getNonRemovedOriginator(id))){
				return false;
			}
			if(((Manual)tree.getNode(id)).getOriginators().size() == conf.getRemovedOriginator(id).size()){
				// we have removed all
				return false;
			}
		}
		return true;
	}

	@SuppressWarnings("unused")
	public List<Pair<Object, Object>> getSubdivisions(int divisions) throws Exception {
		// we have some bit fun here :-D
		List<Pair<Object, Object>> subdivisions = new ArrayList<Pair<Object,Object>>();
		BitSet min;
		BitSet max = new BitSet();
		// easy, we get this one for free :D
		min = (BitSet) currentConfiguration.clone();
		throw new Exception("Not implemented\n");
		//return null;
	}

}
