package org.processmining.processtree.conversion.yawl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.processmining.framework.util.Pair;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.yawlfoundation.yawl.editor.elements.model.Decorator;
import org.yawlfoundation.yawl.editor.elements.model.EngineIdentifier;
import org.yawlfoundation.yawl.editor.elements.model.VertexContainer;
import org.yawlfoundation.yawl.editor.net.NetGraphModel;

public class YawlLoopDefConverter implements YawlConverter{

	private YawlConversionType conversionType = YawlConversionType.PRIVATE;
	private Map<Node, NetGraphModel> subNets;
	private Map<NetGraphModel, Map<Node, Pair<VertexContainer, VertexContainer>>> fragments;

	public YawlLoopDefConverter(YawlConversionType conversionType) {
		this.conversionType = conversionType;
		subNets = new HashMap<Node, NetGraphModel>();
		fragments = new HashMap<NetGraphModel, Map<Node, Pair<VertexContainer, VertexContainer>>>();
	}

	public Pair<VertexContainer, VertexContainer> convert(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		if (conversionType == YawlConversionType.COMPOSITE && pt2y.hasMultipleParents(node)) {
			return convertComposite(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.SHARED && pt2y.hasMultipleParents(node)) {
			return convertShared(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.COMSHARED && pt2y.hasMultipleParents(node)) {
			return convertShared(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		if (conversionType == YawlConversionType.SIMPLE && pt2y.hasMultipleParents(node)) {
			return convertSimple(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
		}
		return convertPrivate(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
	}

	private Pair<VertexContainer, VertexContainer> convertPrivate(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		return convertBasic(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
	}

	private Pair<VertexContainer, VertexContainer> convertComposite(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {

		NetGraphModel subNet;
		if (!subNets.containsKey(node)) {
			subNet = YAWLUtil.createSubnet(node, "LOOP_net" + pt2y.newId());
			
			VertexContainer inputCondition = YAWLUtil.createInputCondition(subNet, pt2y.newId());
			VertexContainer outputCondition = YAWLUtil.createOutputCondition(subNet, pt2y.newId());
			
			Pair<VertexContainer, VertexContainer> fragment = convertBasic(tree, node, subNet, outputCondition, pt2y, mappingEdgeIds);
			YAWLUtil.createEdge(subNet, inputCondition, fragment.getFirst());
			YAWLUtil.createEdge(subNet, fragment.getSecond(), outputCondition);
			
			subNets.put(node, subNet);
		} else {
			subNet = subNets.get(node);
		}
		
		VertexContainer compositeTask = YAWLUtil.createCompositeTask(netGraphModel, node.getName() + node.getID().toString());
		YAWLUtil.setDecomposition(compositeTask, subNet);
		YAWLUtil.setVariables(netGraphModel, compositeTask, node.getReadVariablesRecursive(), node.getWrittenVariablesRecursive(), false);
		
		return new Pair<VertexContainer, VertexContainer>(compositeTask, compositeTask);
	}

	private Pair<VertexContainer, VertexContainer> convertShared(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		Pair<VertexContainer, VertexContainer> fragment;
		
		if (!fragments.containsKey(netGraphModel) || !fragments.get(netGraphModel).containsKey(node)) {
			
			VertexContainer condition = YAWLUtil.createCondition(netGraphModel, "LOOP_condition" + pt2y.newId());
						
			fragment = convertBasic(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
			YAWLUtil.createEdge(netGraphModel, fragment.getSecond(), condition);
			

			if (!fragments.containsKey(netGraphModel)) {
				fragments.put(netGraphModel, new HashMap<Node, Pair<VertexContainer, VertexContainer>>());
			}
			fragments.get(netGraphModel).put(node, fragment);
		} 
		
		fragment = fragments.get(netGraphModel).get(node);
		
		VertexContainer alphaTask = YAWLUtil.createAtomicTask(netGraphModel, "LOOP_alpha" + pt2y.newId());
		YAWLUtil.setSplitDecorator(alphaTask, Decorator.AND_TYPE);

		VertexContainer omegaTask = YAWLUtil.createAtomicTask(netGraphModel, "LOOP_omega" + pt2y.newId());
		YAWLUtil.setJoinDecorator(omegaTask, Decorator.AND_TYPE);
		
		YAWLUtil.createEdge(netGraphModel, alphaTask, omegaTask);
		YAWLUtil.createEdge(netGraphModel, alphaTask, fragment.getFirst());
		YAWLUtil.createEdge(netGraphModel, fragment.getSecond(), omegaTask);
		
		return new Pair<VertexContainer, VertexContainer>(alphaTask, omegaTask);
	}

	private Pair<VertexContainer, VertexContainer> convertSimple(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		
		Pair<VertexContainer, VertexContainer> fragment;
		
		if (!fragments.containsKey(netGraphModel) || !fragments.get(netGraphModel).containsKey(node)) {
			
			fragment = convertPrivate(tree, node, netGraphModel, exitNode, pt2y, mappingEdgeIds);
			YAWLUtil.setSplitDecorator(fragment.getSecond(), Decorator.XOR_TYPE);

			if (!fragments.containsKey(netGraphModel)) {
				fragments.put(netGraphModel, new HashMap<Node, Pair<VertexContainer, VertexContainer>>());
			}
			fragments.get(netGraphModel).put(node, fragment);
		} 
			
		fragment = fragments.get(netGraphModel).get(node);
		
		return fragment;
	}

	private Pair<VertexContainer, VertexContainer> convertBasic(ProcessTree tree, Node node, NetGraphModel netGraphModel,
			VertexContainer exitNode, ProcessTree2Yawl pt2y, Map<Edge, Pair<EngineIdentifier, EngineIdentifier>> mappingEdgeIds) {
		List<Pair<VertexContainer, VertexContainer>> fragments = new ArrayList<Pair<VertexContainer, VertexContainer>>();
		VertexContainer alphaTask = YAWLUtil.createAtomicTask(netGraphModel, "LOOPStart" + pt2y.newId());
		for (Node childNode: ((Block)node).getChildren()) {
			fragments.add(pt2y.convert(tree, childNode, netGraphModel, exitNode, pt2y, mappingEdgeIds));
		}
		if (fragments.size() != 3) {
			return null;
		}
		Pair<VertexContainer, VertexContainer> doFragment = fragments.get(0);
		Pair<VertexContainer, VertexContainer> redoFragment = fragments.get(1);
		Pair<VertexContainer, VertexContainer> exitFragment = fragments.get(2);
		
		YAWLUtil.setJoinDecorator(doFragment.getFirst(), Decorator.XOR_TYPE);
		
		VertexContainer condition = YAWLUtil.createCondition(netGraphModel, "DEF" + pt2y.newId());
		
		Edge doEdge = ((Block)node).getOutgoingEdges().get(0);
		Edge redoEdge = ((Block)node).getOutgoingEdges().get(1);
		Edge exitEdge = ((Block)node).getOutgoingEdges().get(2);
		
		YAWLUtil.createEdge(netGraphModel, alphaTask, doFragment.getFirst(), doEdge.isHideable(), doEdge.isBlockable());
		mappingEdgeIds.put(doEdge, new Pair<EngineIdentifier, EngineIdentifier>(alphaTask.getVertex().getEngineIdentifier(), doFragment.getFirst().getVertex().getEngineIdentifier()));
		YAWLUtil.createEdge(netGraphModel, doFragment.getSecond(), condition);
		YAWLUtil.createEdge(netGraphModel, condition, redoFragment.getFirst(), redoEdge.isHideable(), redoEdge.isBlockable());
		mappingEdgeIds.put(redoEdge, new Pair<EngineIdentifier, EngineIdentifier>(condition.getVertex().getEngineIdentifier(), redoFragment.getFirst().getVertex().getEngineIdentifier()));
		YAWLUtil.createEdge(netGraphModel, condition, exitFragment.getFirst(), exitEdge.isHideable(), exitEdge.isBlockable());
		mappingEdgeIds.put(exitEdge, new Pair<EngineIdentifier, EngineIdentifier>(condition.getVertex().getEngineIdentifier(), exitFragment.getFirst().getVertex().getEngineIdentifier()));
		YAWLUtil.createEdge(netGraphModel, redoFragment.getSecond(), doFragment.getFirst());
		
		return new Pair<VertexContainer, VertexContainer>(alphaTask, exitFragment.getSecond());
	}
}

