package org.processmining.processtree.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Variable;

public abstract class AbstractNode extends ProcessTreeElementImpl implements Node {
	protected ProcessTree tree;
	protected final Set<Variable> readVars;
	protected final Set<Variable> writeVars;
	protected final Set<Variable> remReadVars;
	protected final Set<Variable> remWriteVars;
	protected final List<Edge> incoming;

	public AbstractNode(String name) {
		this(name, new ArrayList<Edge>());
	}
	
	public AbstractNode(UUID id, String name) {
		this(id, name, new ArrayList<Edge>());
	}

	public AbstractNode(String name, List<Edge> incoming) {
		this(UUID.randomUUID(), name, incoming);
	}

	public AbstractNode(UUID id, String name, List<Edge> incoming) {
		super(id, name);
		this.readVars = new HashSet<Variable>();
		this.writeVars = new HashSet<Variable>();
		this.remReadVars = new HashSet<Variable>();
		this.remWriteVars = new HashSet<Variable>();
		this.incoming = incoming;
	}
	
	public AbstractNode(Node n){
		this(n.getID(), n);
	}
	
	public AbstractNode(UUID id, Node n){
		super(id, n);
		this.readVars = new HashSet<Variable>();
		this.writeVars = new HashSet<Variable>();
		this.remReadVars = new HashSet<Variable>();
		this.remWriteVars = new HashSet<Variable>();
		this.incoming = new ArrayList<Edge>();
	}

	public void setProcessTree(ProcessTree tree) {
		this.tree = tree;
	}

	public UUID getID() {
		return id;
	}

	public boolean isRoot() {
		return incoming.isEmpty();
	}

	public Collection<Variable> getReadVariables() {
		return readVars;
	}
	
	public Collection<Variable> getRemovableReadVariables(){
		return remReadVars;
	}

	public Collection<Variable> getWrittenVariables() {
		return writeVars;
	}
	
	public Collection<Variable> getRemovableWrittenVariables(){
		return remWriteVars;
	}

	public boolean addReadVariable(Variable var) {
		return readVars.add(var);
	}
	
	public boolean addRemovableReadVariable(Variable var){
		return remReadVars.add(var);
	}
	
	public boolean removeReadVariable(Variable var) {
		return readVars.remove(var);
	}
	
	public boolean removeRemovableReadVariable(Variable var){
		return remReadVars.remove(var);
	}

	public boolean addWriteVariable(Variable var) {
		return writeVars.add(var);
	}
	
	public boolean addRemovableWriteVariable(Variable var){
		return remWriteVars.add(var);
	}

	public boolean removeWriteVariable(Variable var) {
		return writeVars.remove(var);
	}
	
	public boolean removeRemovableWriteVariable(Variable var){
		return remWriteVars.remove(var);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addParent(org.processmining.processtree
	 * .Block, org.processmining.processtree.Expression)
	 */
	public Edge addParent(UUID id, Block parent, Expression expression) {
		Edge e = new EdgeImpl(id, parent, this, expression);
		tree.addEdge(e);
		addIncomingEdge(e);
		readVars.addAll(expression.getVariables());
		return e;
	}

	public Edge addParent(Block parent, Expression expression) {
		return addParent(UUID.randomUUID(), parent, expression);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addParent(org.processmining.processtree
	 * .Block)
	 */
	public Edge addParent(UUID id, Block parent) {
		return addParent(parent, EdgeImpl.NOEXPRESSION);
	}

	public Edge addParent(Block parent) {
		return addParent(UUID.randomUUID(), parent);
	}
	
	public void addIncomingEdge(Edge edge) {
		if(!incoming.contains(edge)){
			incoming.add(edge);
		}
	}
	
	public boolean removeIncomingEdge(Edge edge){
		return incoming.remove(edge);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#getParents()
	 */
	public Collection<Block> getParents() {
		Block[] parents = new Block[incoming.size()];
		for (int i = incoming.size(); i-- > 0;) {
			parents[i] = incoming.get(i).getSource();
		}
		return Arrays.asList(parents);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#getIncomingEdges()
	 */
	public List<Edge> getIncomingEdges() {
		return Collections.unmodifiableList(incoming);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#numParents()
	 */
	public int numParents() {
		return incoming.size();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Node#getName()
	 */
	public String getName() {
		return name;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Node#getProcessTree()
	 */
	public ProcessTree getProcessTree() {
		return tree;
	}

	public String toString() {
		return tree.toString(this);
	}
}
