package org.processmining.processtree.impl;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.hash.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;
import java.util.UUID;

import org.processmining.processtree.Block;
import org.processmining.processtree.Block.And;
import org.processmining.processtree.Block.Def;
import org.processmining.processtree.Block.DefLoop;
import org.processmining.processtree.Block.Or;
import org.processmining.processtree.Block.PlaceHolder;
import org.processmining.processtree.Block.Seq;
import org.processmining.processtree.Block.Xor;
import org.processmining.processtree.Block.XorLoop;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Event.Message;
import org.processmining.processtree.Event.TimeOut;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.Originator;
import org.processmining.processtree.Originator.Group;
import org.processmining.processtree.Originator.Resource;
import org.processmining.processtree.Originator.Role;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTreeElement;
import org.processmining.processtree.Task;
import org.processmining.processtree.Task.Automatic;
import org.processmining.processtree.Task.Manual;
import org.processmining.processtree.Variable;

public class ProcessTreeImpl extends ProcessTreeElementImpl implements ProcessTree {

	protected final TObjectIntMap<Node> nodes = new TObjectIntHashMap<Node>();
	protected final Set<Edge> edges = new HashSet<Edge>();
	protected final Set<Variable> variables = new HashSet<Variable>();
	protected final List<Originator> originators = new ArrayList<Originator>();
	protected final Set<Expression> expressions = new HashSet<Expression>();
	protected Node root;
	private int startIndex = 0;

	public ProcessTreeImpl(String name) {
		this(UUID.randomUUID(), name);
	}

	public ProcessTreeImpl() {
		this(UUID.randomUUID());
	}
	
	public ProcessTreeImpl(UUID id){
		this(id, id.toString());
	}
	
	public ProcessTreeImpl(UUID id, String name){
		super(id, name);
	}
	
	public ProcessTreeImpl(ProcessTree tree){
		super(tree);
		HashSet<Manual> manualTasks = new HashSet<Manual>();
		HashSet<Block> blocks = new HashSet<Block>();
		for(Node n: tree.getNodes()){
			Node np = null;
			if(n instanceof Seq){
				np = new AbstractBlock.Seq((Seq)n);
			}
			if(n instanceof And){
				np = new AbstractBlock.And((And)n);
			}
			if(n instanceof Xor){
				np = new AbstractBlock.Xor((Xor)n);
			}
			if(n instanceof Def){
				np = new AbstractBlock.Def((Def)n);
			}
			if(n instanceof Or){
				np = new AbstractBlock.Or((Or)n);
			}
			if(n instanceof XorLoop){
				np = new AbstractBlock.XorLoop((XorLoop)n);
			}
			if(n instanceof DefLoop){
				np = new AbstractBlock.DefLoop((DefLoop)n);
			}
			if(n instanceof PlaceHolder){
				np = new AbstractBlock.PlaceHolder((PlaceHolder)n);
			}
			if(n instanceof Message){
				np = new AbstractEvent.Message((Message)n);
			}
			if(n instanceof TimeOut){
				np = new AbstractEvent.TimeOut((TimeOut)n);
			}
			if(n instanceof Block){
				blocks.add((Block) n);
			}
			if(n instanceof Manual){
				np = new AbstractTask.Manual((Manual)n);
				manualTasks.add((Manual)n);
			}
			if(n instanceof Automatic){
				np = new AbstractTask.Automatic((Automatic)n);
			}
			this.addNode(np);
			np.setProcessTree(this);
		}
		for(Originator org: tree.getOriginators()){
			Originator orgp = null;
			if(org instanceof Group){
				orgp = new AbstractOriginator.Group((Group)org);
			}
			if(org instanceof Resource){
				orgp = new AbstractOriginator.Resource((Resource)org);
			}
			if(org instanceof Role){
				orgp = new AbstractOriginator.Role((Role)org);
			}
			this.addOriginator(orgp);
		}
		for(Variable var: tree.getVariables()){
			Variable varp = new VariableImpl(var);
			this.addVariable(varp);
		}
		// set the originators
		for(Manual m: manualTasks){
			Collection<Originator> orgs = new HashSet<Originator>();
			Collection<Originator> remOrgs = new HashSet<Originator>();
			for(Originator org: m.getOriginators()){
				orgs.add(this.getOriginator(org.getID()));
			}
			if(m.getRemovableOriginators() != null){
				for(Originator org: m.getRemovableOriginators()){
					remOrgs.add(this.getOriginator(org.getID()));
				}
			}
			((Manual)this.getNode(m.getID())).addOriginators(orgs);
			((Manual)this.getNode(m.getID())).addRemovableOriginators(remOrgs);
		}
		// set the variables
		for(Node n: tree.getNodes()){
			for(Variable var: n.getReadVariables()){
				this.getNode(n.getID()).getReadVariables().add(this.getVariable(var.getID()));
			}
			for(Variable var: n.getWrittenVariables()){
				this.getNode(n.getID()).getWrittenVariables().add(this.getVariable(var.getID()));
			}
			for(Variable var: n.getRemovableReadVariables()){
				this.getNode(n.getID()).getRemovableReadVariables().add(this.getVariable(var.getID()));
			}
			for(Variable var: n.getRemovableWrittenVariables()){
				this.getNode(n.getID()).getRemovableWrittenVariables().add(this.getVariable(var.getID()));
			}
			
		}
		// set the expressions
		for(Expression expr: tree.getExpressions()){
			Expression exprp = new ExpressionImpl(expr);
			for(Variable var: expr.getVariables()){
				exprp.addVariable(this.getVariable(var.getID()));
			}
			this.addExpression(exprp);
		}
		
		// set the edges
		for(Block b: blocks){
			for(Edge e: b.getOutgoingEdges()){
				Expression expr = EdgeImpl.NOEXPRESSION;
				if(e.hasExpression()){
					expr = this.getExpression(e.getExpression().getID());
				}
				assert e.getSource().getID().equals(b.getID());
				Edge ep = new EdgeImpl(e, (Block)this.getNode(e.getSource().getID()), this.getNode(e.getTarget().getID()), expr);
				((Block)this.getNode(b.getID())).addOutgoingEdge(ep);
				this.getNode(e.getTarget().getID()).addIncomingEdge(ep);
				this.addEdge(ep);
				for(Expression ex: e.getExpressions()){
					Expression exp = this.getExpression(ex.getID());
					ep.getExpressions().add(exp);
					if(e.getRemovableExpressions().contains(ex)){
						ep.getRemovableExpressions().add(exp);
					}
				}
			}
		}
		if(tree.getRoot() != null){
			this.root = this.getNode(tree.getRoot().getID());
		}
		else{
			this.root = null;
		}
	}

	public void setRoot(Node root) {
		this.root = root;
	}

	public boolean isTree() {
		return isTree(root, new boolean[startIndex]);
	}

	private boolean isTree(Node node, boolean[] visisted) {
		int nodeIndex = nodes.get(node);
		if (visisted[nodeIndex] || node.getParents().size() > 1) {
			return false;
		} else {
			visisted[nodeIndex] = true;
			if (!(node instanceof Block)) {
				return true;
			} else {
				Block b = (Block) node;
				for (Node n : b.getChildren()) {
					if (!isTree(n, visisted)) {
						return false;
					}
				}
				return true;
			}
		}
	}

	public boolean isDag() {
		boolean[][] path = new boolean[startIndex][startIndex];
		return isDag(root, path);
	}

	private boolean isDag(Node node, boolean[][] path) {
		int from = nodes.get(node);

		if (node instanceof Block) {
			Block block = (Block) node;
			for (Node n : block.getChildren()) {
				int to = nodes.get(n);
				if (path[to][from]) {
					return false;
				}
				path[from][to] = true;
				for (int i = 0; i < startIndex; i++) {
					if (path[i][from]) {
						path[i][to] = true;
					}
					if (!isDag(n, path)) {
						return false;
					}
				}
			}

		} else {
			path[from][from] = true;
		}
		return true;

	}
	
	public ProcessTree toTree() {
		ProcessTree tree = new ProcessTreeImpl(this);
		// now just transform the tree to a tree
		if(tree.isTree()){
			return tree;
		}
		else{
			Stack<Node> nonDagNodes = new Stack<Node>();
			for(Node n: tree.getNodes()){
				if(n.getIncomingEdges().size() > 1){
					nonDagNodes.add(n);
				}
			}
			while(!nonDagNodes.empty()){
				Node n = nonDagNodes.pop();
				for(int i = 1; i < n.getIncomingEdges().size(); i++){
					Node newNode = null;
					if(n instanceof Automatic){
						newNode = new AbstractTask.Automatic(UUID.randomUUID(), (Task.Automatic) n);
					}
					if(n instanceof Manual){
						newNode = new AbstractTask.Manual(UUID.randomUUID(), n.getName(), ((Manual) n).getOriginators());
					}
					if(n instanceof Seq){
						newNode = new AbstractBlock.Seq(UUID.randomUUID(), n.getName());
					}
					if(n instanceof And){
						newNode = new AbstractBlock.And(UUID.randomUUID(), n.getName());
					}
					if(n instanceof Xor){
						newNode = new AbstractBlock.Xor(UUID.randomUUID(), n.getName());
					}
					if(n instanceof Or){
						newNode = new AbstractBlock.Or(UUID.randomUUID(), n.getName());
					}
					if(n instanceof Def){
						newNode = new AbstractBlock.Def(UUID.randomUUID(), n.getName());
					}
					if(n instanceof TimeOut){
						newNode = new AbstractEvent.TimeOut(n.getName(), ((TimeOut) n).getMessage());
					}
					if(n instanceof Message){
						newNode = new AbstractEvent.Message(n.getName(), ((Message) n).getMessage());
					}
					if(n instanceof XorLoop){
						newNode = new AbstractBlock.XorLoop(UUID.randomUUID(), n.getName());
					}
					if(n instanceof DefLoop){
						newNode = new AbstractBlock.DefLoop(UUID.randomUUID(), n.getName());
					}
					if(n instanceof PlaceHolder){
						newNode = new AbstractBlock.PlaceHolder(UUID.randomUUID(), n.getName());
					}
					tree.addNode(newNode);
					newNode.setProcessTree(tree);
					newNode.addIncomingEdge(n.getIncomingEdges().get(i));
					n.getIncomingEdges().get(i).setTarget(newNode);
					// check if we have a block, we need to copy the children
					if(n instanceof Block){
						for(Edge childEdge: ((Block) n).getOutgoingEdges()){
							Edge edge = ((Block)newNode).addChild(childEdge.getTarget());
							edge.setBlockable(childEdge.isBlockable());
							edge.setHideable(childEdge.isHideable());
							tree.addEdge(edge);
							// add the child to the stack of nonDagNodes
							nonDagNodes.push(childEdge.getTarget());
						}
					}
				}
				while(n.getIncomingEdges().size() > 1){
					n.removeIncomingEdge(n.getIncomingEdges().get(1));
				}
			}
		}
		return tree;
	}
	
	public boolean isConfigurable(){
		for(Edge edge: getEdges()){
			if(edge.isHideable() || edge.isBlockable()){
				return true;
			}
		}
		for(Node node: getNodes()){
			if(node instanceof PlaceHolder){
				return true;
			}
		}
		return false;
	}

	@Override
	public Node getRoot() {
		return root;
	}

	@Override
	public Collection<Node> getNodes() {
		return Collections.unmodifiableCollection(nodes.keySet());
	}

	@Override
	public Collection<Edge> getEdges() {
		return Collections.unmodifiableCollection(edges);
	}

	@Override
	public Collection<Variable> getVariables() {
		return Collections.unmodifiableCollection(variables);
	}
	
	@Override
	public Collection<Expression> getExpressions() {
		/*
		HashSet<Expression> expr = new HashSet<Expression>();
		for(Edge e: getEdges()){
			if(e.hasExpression()){
				expr.add(e.getExpression());
			}
		}
		return Collections.unmodifiableCollection(expr);
		*/
		return Collections.unmodifiableCollection(expressions);
	}

	@Override
	public boolean addNode(Node node) {
		if (nodes.containsKey(node)) {
			return false;
		}
		nodes.put(node, (startIndex++));
		node.setProcessTree(this);
		return true;
	}
	
	public boolean removeNode(Node node){
		if(nodes.containsKey(node)){
			nodes.remove(node);
			return true;
		}
		return false;
	}

	@Override
	public boolean addVariable(Variable variable) {
		return variables.add(variable);
	}
	
	public boolean removeVariable(Variable variable){
		// update all the nodes which could use this variable
		for(Node n: this.getNodes()){
			n.removeReadVariable(variable);
			n.removeWriteVariable(variable);
			n.removeRemovableReadVariable(variable);
			n.removeRemovableWriteVariable(variable);
		}
		// also do the expressions (note that expression is ill defined now,
		// but it would be also with keeping the variable there)
		for(Expression ex: this.getExpressions()){
			ex.removeVariable(variable);
		}
		return variables.remove(variable);
	}

	@Override
	public boolean addEdge(Edge edge) {
		return edges.add(edge);
	}
	
	public boolean removeEdge(Edge edge){
		return edges.remove(edge);
	}
	
	public boolean addExpression(Expression expression){
		return expressions.add(expression);
	}
	
	public boolean removeExpression(Expression expression){
		return expressions.remove(expression);
	}

	@Override
	public <N extends Node> Iterator<N> iterator(final Type type) {

		return new Iterator<N>() {

			private final Iterator<Node> it = nodes.keySet().iterator();

			private N next = null;

			@SuppressWarnings("unchecked")
			@Override
			public boolean hasNext() {
				if (next == null) {
					while (it.hasNext()) {
						Node n = it.next();
						if (Type.getTypeOf(n).equals(type)) {
							next = (N) n;
							return true;
						}
					}
					return false;
				} else {
					return true;
				}
			}

			@Override
			public N next() {
				N n = next;
				hasNext();
				return n;
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException("Remove node supported");
			}
		};
	}

/*
	public boolean equals(Object o) {
		return (o instanceof ProcessTree ? ((ProcessTree) o).getRoot().equals(root) : false);
	}
*/

/*
	public int hashCode() {
		return root.hashCode();
	}
*/

	public String toString() {
		return toString(root);
	}

	public String toString(Node node) {
		StringBuilder builder = new StringBuilder();
		toString(node, builder);
		return builder.toString();
	}

	private void toString(Node node, StringBuilder builder) {
		if (node != null) {
			builder.append(node.toStringShort());
			if (node instanceof Block) {
				Block block = (Block) node;
				builder.append("(");
				Iterator<Edge> it = block.getOutgoingEdges().iterator();
				while (it.hasNext()) {
					Edge edge = it.next();
					builder.append(edge.getExpression());
					toString(edge.getTarget(), builder);
					if (it.hasNext()) {
						builder.append(", ");
					}
				}
				builder.append(")");
			}
		}
	}

	public Type getType(Node node) {
		return Type.getTypeOf(node);
	}

	public int size() {
		return nodes.size();
	}

	public Collection<Originator> getOriginators() {
		return Collections.unmodifiableCollection(originators);
	}

	public boolean addOriginator(Originator originator) {
		return originators.add(originator);
	}
	
	public boolean removeOriginator(Originator originator) {
		// also remove the originators from the various tasks
		for(Node n: this.getNodes()){
			if(n instanceof Manual){
				((Manual)n).removeOriginator(originator);
			}
		}
		return originators.remove(originator);
	}

	public Node getNode(UUID id) {
		for(Node n: getNodes()){
			if(n.getID().equals(id)){
				return n;
			}
		}
		return null;
	}

	public Originator getOriginator(UUID id) {
		for(Originator o: getOriginators()){
			if(o.getID().equals(id)){
				return o;
			}
		}
		return null;
	}

	public Edge getEdge(UUID id) {
		for(Edge e: getEdges()){
			if(e.getID().equals(id)){
				return e;
			}
		}
		return null;
	}
	
	public Variable getVariable(UUID id){
		for(Variable v: getVariables()){
			if(v.getID().equals(id)){
				return v;
			}
		}
		return null;
	}
	
	public Expression getExpression(UUID id){
		for(Expression expr: getExpressions()){
			if(expr.getID().equals(id)){
				return expr;
			}
		}
		return null;
	}
	public Collection<ProcessTreeElement> getElements() {
		HashSet<ProcessTreeElement> ret = new HashSet<ProcessTreeElement>();
		ret.add(this);
		ret.addAll(this.getNodes());
		ret.addAll(this.getEdges());
		ret.addAll(this.getOriginators());
		ret.addAll(this.getVariables());
		ret.addAll(this.getExpressions());
		return ret;
	}
	
	public ProcessTreeElement getElement(UUID id){
		if(this.getID().equals(id)){
			return this;
		}
		if(getNode(id) != null){
			return getNode(id);
		}
		if(getOriginator(id) != null){
			return getOriginator(id);
		}
		if(getEdge(id) != null){
			return getEdge(id);
		}
		if(getVariable(id) != null){
			return getVariable(id);
		}
		if(getExpression(id) != null){
			return getExpression(id);
		}
		return null;
	}
}
