package org.processmining.processtree.ptml;

import java.util.Map;
import java.util.UUID;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.impl.EdgeImpl;
import org.xmlpull.v1.XmlPullParser;

public class PtmlIsParent extends PtmlEdge {

	public final static String TAG = "parentsNode";
	
	private UUID expression;
	private boolean blockable;
	private boolean hideable;
	protected PtmlPropertyIndependent propertyIndependent;
	protected PtmlPropertyDependent propertyDependent;

	public PtmlIsParent(UUID id) {
		super(TAG, id);
		expression = null;
		blockable = false;
		hideable = false;
	}

	protected void importAttributes(XmlPullParser xpp, Ptml ptml) {
		super.importAttributes(xpp, ptml);
		importExpression(xpp, ptml);
		importBlockable(xpp, ptml);
		importHideable(xpp, ptml);
	}

	private void importExpression(XmlPullParser xpp, Ptml ptml) {
		String value = xpp.getAttributeValue(null, "expressionId");
		if (value != null) {
			//expression = new String(value);
			expression = UUID.fromString(value);
		}
		ptml.log(expression != null ? value : "<no expression>");
	}

	private void importBlockable(XmlPullParser xpp, Ptml ptml) {
		String value = xpp.getAttributeValue(null, "blockable");
		if (value != null) {
			blockable = new Boolean(value);
		}
		ptml.log(value != null ? value : "<no value>");
	}

	private void importHideable(XmlPullParser xpp, Ptml ptml) {
		String value = xpp.getAttributeValue(null, "hideable");
		if (value != null) {
			hideable = new Boolean(value);
		}
		ptml.log(value != null ? value : "<no value>");
	}

	protected String exportAttributes(Ptml ptml) {
		return super.exportAttributes(ptml) + exportExpression(ptml) + exportBlockable(ptml) + exportHideable(ptml);
	}

	private String exportExpression(Ptml ptml) {
		if (expression != null) {
			return exportAttribute("expressionId", expression.toString(), ptml);
		}
		return "";
	}
	
	private String exportBlockable(Ptml ptml) {
		if (blockable) {
			return exportAttribute("blockable", "true", ptml);
		}
		return "";
	}
	
	private String exportHideable(Ptml ptml) {
		if (hideable) {
			return exportAttribute("hideable", "true", ptml);
		}
		return "";
	}
	
	protected String exportElements(Ptml ptml){
		String s = super.exportElements(ptml);
		// property export
		if(propertyIndependent != null){
			s += propertyIndependent.exportElement(ptml);
		}
		if(propertyDependent != null){
			s += propertyDependent.exportElement(ptml);
		}
		return s;
	}
	
	protected boolean importElements(PluginContext context, XmlPullParser xpp, Ptml ptml) {
		PtmlPropertyIndependent propertyInd = null;
		if(xpp.getName().equals(PtmlPropertyIndependent.TAG)){
			propertyInd = new PtmlPropertyIndependent();
		}
		if(propertyInd != null){
			propertyIndependent = propertyInd;
			propertyIndependent.importElement(context, xpp, ptml);
			return true;
		}
		
		PtmlPropertyDependent propertyDep = null;
		if(xpp.getName().equals(PtmlPropertyDependent.TAG)){
			propertyDep = new PtmlPropertyDependent();
		}
		if(propertyDep != null){
			propertyDependent = propertyDep;
			propertyDependent.importElement(context, xpp, ptml);
			return true;
		}
		return false;
	}
	
	public void unmarshall(ProcessTree tree, Map<UUID, Node> nodeMap, Map<UUID, Edge> edgeMap, Map<UUID, Expression> exprMap) {
		Node sourceNode = nodeMap.get(sourceId);
		Node targetNode = nodeMap.get(targetId);
		if (sourceNode instanceof Block) {
			Expression exp = EdgeImpl.NOEXPRESSION;
			if (expression != null) {
				exp = exprMap.get(expression);
			}
			Edge edge = targetNode.addParent(id, (Block) sourceNode, exp);
			edge.setBlockable(blockable);
			edge.setHideable(hideable);
			((Block) sourceNode).addOutgoingEdge(edge);
			edgeMap.put(id, edge);
			if(propertyIndependent != null){
				propertyIndependent.unmarshall(edge);
			}
			if(propertyDependent != null){
				propertyDependent.unmarshall(edge);
			}
		}
	}
	
	public void marshall(ProcessTree tree, Edge edge, Map<Edge, UUID> edgeMap) {
		id = edge.getID();
		sourceId = edge.getSource().getID();
		targetId = edge.getTarget().getID();
		expression = null;
		if (edge.hasExpression()) {
			expression = edge.getExpression().getID();
		}
		blockable = edge.isBlockable();
		hideable = edge.isHideable();
		edgeMap.put(edge, id);
		if(!edge.getIndependentProperties().isEmpty()){
			propertyIndependent = new PtmlPropertyIndependent();
			propertyIndependent.marshall(edge.getIndependentProperties());
		}
		if(!edge.getDependentProperties().isEmpty()){
			propertyDependent = new PtmlPropertyDependent();
			propertyDependent.marshall(edge.getDependentProperties());
		}
	}
}
